! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: cheese_mod
!
! Authors:  Brendan Perry - NASA/GSFC, brendan.h.perry@nasa.gov
!           K.D.Kuntz (NASA/GSFC) and The Johns Hopkins University, kuntz@pha.jhu.edu
!           Steven L. Snowden - NASA/GSFC (original perl algorithm)
! 
! Description: Extended Source Analysis Software
!
! Comments:
!
! $Id: cheese_mod.f90,v 1.3 2009/02/26 16:39:45 aibarra Exp $
!-
!
!
! This task creates "cheese" masks after running source detection on full-field
! images.  It uses the SAS tasks emask, region, and edetect_chain.
!
!  Usage:
!
module cheese_mod

  use dal
  use cal
  use param
  use dataFilter
  use errorHandling
  use sasdal
  use gofio_mod

  implicit none

  private

  character(len=*), parameter :: &
       mos1filepar  = "mos1file", &    ! input EMOS event lists from espfilt
       mos2filepar  = "mos2file", &    ! input EMOS event lists from espfilt
       pnfilepar    = "pnfile", &      ! input EPN event lists from espfilt
       pnootpar     = "pnootfile", &   ! input EPN out of time event list
       scalepar     = "scale", &       ! E fraction (exclusion radius for pt srcs)
       rateTpar     = "ratetotal", &   ! Flux threshold for exclusion of point sources
       rateSpar     = "ratesoft", &    ! Flux threshold for excl of pnt srcs (soft)
       rateHpar     = "ratehard", &    ! Flux threshold for excl of pnt srcs (hard)
       distpar      = "dist", &        ! Min sep (arcsecs) between masked sources
       mlminpar     = "mlmin", &       ! Min Max Likelihood detection value
       elowpar      = "elowlist", &    ! Low energy limit for the bands (eV)
       ehighpar     = "ehighlist", &   ! High energy limit for the bands (eV)
       keepspar     = "keepinterfiles" ! keep (delete) intermediary files?

  public cheese

contains

  !---------------------------------------------------------------------
  subroutine cheese
    !---------------------------------------------------------------------

    type(DataSetT) :: inOBJIMG_S, inOBJSPEC_S, outSPdetmap_S, outSPskymap_S, outOBJIMG_S
    type(ArrayT)   :: inOBJIMG_A, outSPmap_A, out_A
    type(TableT)   :: inOBJSPEC_T, inOBJIMG_T
    character (len=2) :: inst, instO, instS, cc, bandd
    character (len=4) :: mos1ecf, mos2ecf, pnecf
    character (len=5) :: expid, any
    character (len=6) :: celow(2), cehigh(2)
    character (len=10) :: obsid, obsidO, obsidS, ctype(2), cunit(2), radecsys
    character (len=12) :: str_crval1, str_crval2, str_detx, str_dety
    character (len=18) :: prefix, type, suffix, logfile
    character (len=25) :: smode, smodeO, smodeS, submode
    character (len=1024) :: inmos1_F, inmos2_F, inpn_F, inpnoot_F, &
         outM1fovT_F, outM1fovS_F, outM1fovH_F, outM2fovT_F, outM2fovS_F,&
         outM2fovH_F, outPNfovT_F, outPNfovS_F, outPNfovH_F
    character (len=1024) :: atthk_F, detmask_F, expimg_F, regiondet_F, regionsky_F, &
         cheese_F, piminlist, pimaxlist, ecflist
    character (len=4096) :: command
    character (len=1024) :: imglist, evtlist
    integer, dimension(2) :: outIMGdim
    integer :: numMOS1ins, numMOS2ins, numPNins
    integer :: chansobs, instnum, typeF, nRowsO, nRowsS, revolut, revolutO, &
         revolutS, spectrol, inChans, spXsize, spYsize, ff
    integer :: elow(2), ehigh(2)
    integer :: i, j, k, n, nEVLIs
    integer :: status = 0, nBandsL, nBandsH
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: outSPmap_IP, aDI32, &
         inOBJIMG_IP
    integer, parameter :: specchans = 2400
    integer, parameter :: xGood = 780, yGood = 780
    logical :: warndie, die, go, mos1, mos2, pn, keep
    real(single), dimension(:,:), pointer :: outSPmap_RP, flaremap_RP, aDR32
    real(double), dimension(:,:), pointer :: aDR64
    real(single), dimension(:,:), pointer :: temp_real32
    real :: tempp, dist, mlmin
    real :: eh, el, exposen, rn, scale, rateT, rateS, rateH, temp
    real(kind=double) :: crval(2), cdelt(2), crpix(2), crota(2), &
         dxcen, dycen, ra_pnt, dec_pnt, pa_pnt
    real :: backO, areaO, expoO, backS, areaS, expoS
    mos1ecf = '1.2' ! parameterize these???!!!
    mos2ecf = '1.2'
    pnecf   = '3.2'
    mos1 = .FALSE.
    mos2 = .FALSE.
    pn   = .FALSE.
    !OLDNEW: mos1S001-clean.fits P0402530201M1S001-objevlifilt.FIT/mos1S001-allevc.fits (espfilt)
    !OLDNEW: mos2S002-clean.fits P0402530201M2S002-objevlifilt.FIT/mos2S002-allevc.fits (espfilt)
    !OLDNEW: pnS003-clean.fits P0402530201PNS003-objevlifilt.FIT/pnS003-allevc.fits (espfilt)
    inmos1_F  = stringParameter(mos1filepar) 
    inmos2_F  = stringParameter(mos2filepar)
    inpn_F    = stringParameter(pnfilepar)
    inpnoot_F = stringParameter(pnootpar)
    if (inmos1_F .ne. "") mos1 = .TRUE.
    if (inmos2_F .ne. "") mos2 = .TRUE.
    if (inpn_F   .ne. "")   pn = .TRUE.
    if ((.not. mos1) .and. (.not. mos2) .and. (.not. pn))then
       call error("NoInsts","At least one of mos1/mos2/pn file must be entered")
    endif
    !
    nBandsL = parameterCount(elowpar)
    nBandsH = parameterCount(ehighpar)
    celow  = ''
    cehigh = ''
    if (nBandsL .gt. 2) then
       write(errstr,*) &
            "Number of elow values must be 1 or 2, exiting..."
       call error("eLowsGT2",trim(errstr))
    endif
    if (nBandsH .gt. 2) then
       write(errstr,*) &
            "Number of ehigh values must be 1 or 2, exiting..."
       call error("eHighsGT2",trim(errstr))
    endif
    if (nBandsL .ne. nBandsH) then
       write(errstr,*) &
            "Number of elow and ehigh values must be equal, exiting..."
       call error("eLowsNEeHighs",trim(errstr))
    else    
       do i=1,nBandsL
          elow(i)  = intParameter(elowpar,i-1)
          ehigh(i) = intParameter(ehighpar,i-1)
          write(celow(i),'(I5)') elow(i)
          write(cehigh(i),'(I5)') ehigh(i)
       end do
    end if
    if (nBandsL .eq. 1) then ! to make single "t" outputs
       celow(2)  = celow(1)
       cehigh(2) = cehigh(1)
    endif
    !
    scale = realParameter(scalepar)
    rateT = realParameter(rateTpar)
    rateS = realParameter(rateSpar)
    rateH = realParameter(rateHpar)
    dist  = realParameter(distpar)
    mlmin = realParameter(mlminpar)
    keep  = booleanParameter(keepspar)
    call printParams(inmos1_F,inmos2_F,inpn_F,inpnoot_F,scale,rateT,rateS,rateH,dist,mlmin,keep)
    !
    imglist   = ""
    evtlist   = ""
    piminlist = ""
    pimaxlist = ""
    ecflist   = ""
    any       = ""
    !
    ! take infiles from params and make FOV images in one loop
    ! keep filenames in array for looping later
    ! build command line params for edetect_chain, emask, region, make_mask
    ! ADD CHECK_CLOBBER???!!!
    write(errstr,*) "Create filtered FOV image files, and edetect_chain parameters:"
    call message(AppMsg,Always)
    ! EMOS1
    if (mos1) then
       ! total band always
       ! OLDNEW: mos1S001-obj-im.fits/mos1S001-fovimt.fits
       call makeOutName(inmos1_F,outM1fovT_F,"t",1,any) ! mos1S001-fovimt.fits
       call makeImFile(inmos1_F,outM1fovT_F,'M1',celow(1),cehigh(2))
       !
       if (nBandsL .eq. 1) then
          imglist   = trim(adjustl(imglist)) // " " // trim(adjustl(outM1fovT_F))
          evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inmos1_F))
          piminlist = trim(adjustl(piminlist)) // " " // trim(adjustl(celow(1)))
          pimaxlist = trim(adjustl(pimaxlist)) // " " // trim(adjustl(cehigh(2)))
          ecflist   = trim(adjustl(ecflist)) // " " // trim(adjustl(mos1ecf))
       else
          ! soft band 
          ! OLDNEW: mos1S001-obj-im-s.fits mos1S001-fovims.fits
          call makeOutName(inmos1_F,outM1fovS_F,"s",1,any) ! mos1S001-fovims.fits
          call makeImFile(inmos1_F,outM1fovS_F,'M1',celow(1),cehigh(1))
          ! hard band 
          ! OLDNEW: mos1S001-obj-im-h.fits mos1S001-fovimh.fits
          call makeOutName(inmos1_F,outM1fovH_F,"h",1,any) ! mos1S001-fovimh.fits
          call makeImFile(inmos1_F,outM1fovH_F,'M1',celow(2),cehigh(2))
          imglist   = trim(adjustl(imglist)) // " " // &
               trim(adjustl(outM1fovT_F)) // " " // &
               trim(adjustl(outM1fovS_F)) // " " // &
               trim(adjustl(outM1fovH_F)) // " " 
          evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inmos1_F))
          piminlist = trim(adjustl(piminlist)) // " " // &
               trim(adjustl(celow(1))) // " " // trim(adjustl(celow(1))) // " " // &
               trim(adjustl(celow(2)))
          pimaxlist = trim(adjustl(pimaxlist)) // " " // &
               trim(adjustl(cehigh(2))) // " " // &
               trim(adjustl(cehigh(1))) // " " // trim(adjustl(cehigh(2)))
          ecflist   = trim(adjustl(ecflist)) // " " // trim(adjustl(mos1ecf)) // &
               " " // trim(adjustl(mos1ecf)) // " " // trim(adjustl(mos1ecf))
       end if
    end if
    ! EMOS2
    if (mos2) then
       ! total band first
       call makeOutName(inmos2_F,outM2fovT_F,"t",1,any) ! mos2S001-fovimt.fits
       call makeImFile(inmos2_F,outM2fovT_F,'M2',celow(1),cehigh(2))
       if (nBandsL .eq. 1) then
          imglist   = trim(adjustl(imglist)) // " " // trim(adjustl(outM2fovT_F))
          evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inmos2_F))
          piminlist = trim(adjustl(piminlist)) // " " // trim(adjustl(celow(1)))
          pimaxlist = trim(adjustl(pimaxlist)) // " " // trim(adjustl(cehigh(2)))
          ecflist   = trim(adjustl(ecflist)) // " " // trim(adjustl(mos2ecf))
       else
          ! soft band
          call makeOutName(inmos2_F,outM2fovS_F,"s",1,any) ! mos2S002-fovims.fits
          call makeImFile(inmos2_F,outM2fovS_F,'M2',celow(1),cehigh(1))
          ! hard band
          call makeOutName(inmos2_F,outM2fovH_F,"h",1,any) ! mos2S002-fovimh.fits
          call makeImFile(inmos2_F,outM2fovH_F,'M2',celow(2),cehigh(2))
          imglist   = trim(adjustl(imglist)) // " " // &
               trim(adjustl(outM2fovT_F)) // " " // &
               trim(adjustl(outM2fovS_F)) // " " // &
               trim(adjustl(outM2fovH_F)) // " " 
          evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inmos2_F))
          piminlist = trim(adjustl(piminlist)) // " " // &
               trim(adjustl(celow(1))) // " " // trim(adjustl(celow(1))) // " " // &
               trim(adjustl(celow(2)))
          pimaxlist = trim(adjustl(pimaxlist)) // " " // &
               trim(adjustl(cehigh(2))) // " " // &
               trim(adjustl(cehigh(1))) // " " // trim(adjustl(cehigh(2)))
          ecflist   = trim(adjustl(ecflist)) // " " // trim(adjustl(mos2ecf)) // &
               " " // trim(adjustl(mos2ecf)) // " " // trim(adjustl(mos2ecf))
       end if
    end if
    ! EPN
    if (pn) then
       ! total band first
       call makeOutName(inpn_F,outPNfovT_F,"t",1,any) ! pnS001-fovimt.fits
       call makeImFile(inpn_F,outPNfovT_F,'PN',celow(1),cehigh(2))
       if (nBandsL .eq. 1) then
          imglist   = trim(adjustl(imglist)) // " " // trim(adjustl(outPNfovT_F))
          evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inpn_F))
          piminlist = trim(adjustl(piminlist)) // " " // trim(adjustl(celow(1)))
          pimaxlist = trim(adjustl(pimaxlist)) // " " // trim(adjustl(cehigh(2)))
          ecflist   = trim(adjustl(ecflist)) // " " // trim(adjustl(pnecf))
       else
          ! soft band
          call makeOutName(inpn_F,outPNfovS_F,"s",1,any) ! pnS001-fovims.fits
          call makeImFile(inpn_F,outPNfovS_F,'PN',celow(1),cehigh(1))
          ! hard band
          call makeOutName(inpn_F,outPNfovH_F,"h",1,any) ! pnS001-fovimh.fits
          call makeImFile(inpn_F,outPNfovH_F,'PN',celow(2),cehigh(2))
          imglist   = trim(adjustl(imglist)) // " " // &
               trim(adjustl(outPNfovT_F)) // " " // &
               trim(adjustl(outPNfovS_F)) // " " // &
               trim(adjustl(outPNfovH_F)) // " " 
          evtlist   = trim(adjustl(evtlist)) // " " // trim(adjustl(inpn_F))
          piminlist = trim(adjustl(piminlist)) // " " // &
               trim(adjustl(celow(1))) // " " // trim(adjustl(celow(1))) // " " // &
               trim(adjustl(celow(2)))
          pimaxlist = trim(adjustl(pimaxlist)) // " " // &
               trim(adjustl(cehigh(2))) // " " // &
               trim(adjustl(cehigh(1))) // " " // trim(adjustl(cehigh(2)))
          ecflist   = trim(adjustl(ecflist)) // " " // trim(adjustl(pnecf)) // &
               " " // trim(adjustl(pnecf)) // " " // trim(adjustl(pnecf))
       end if
    end if
    ! clear out superfluous white space (though edetect_chain works anyway)???!!!
    write(errstr,*) "evtlist   = ",trim(adjustl(evtlist))
    call message(AppMsg,Always)
    write(errstr,*) "imglist   = ",trim(adjustl(imglist))
    call message(AppMsg,Always)
    write(errstr,*) "piminlist = ",trim(adjustl(piminlist))
    call message(AppMsg,Always)
    write(errstr,*) "pimaxlist = ",trim(adjustl(pimaxlist))
    call message(AppMsg,Always)
    write(errstr,*) "ecflist   = ",trim(adjustl(ecflist))
    call message(AppMsg,Always)
    !
    ! check for atthk.fits, else run atthkgen to make attitude file
    ! OLDNEW: atthk.fits atthk.fits
    die = .FALSE.
    go  = .FALSE. 
    atthk_F = 'atthk.fits'
    call chuck_clobber(trim(adjustl(atthk_F)),die,go)
    if (go) then 
       command = "atthkgen atthkset="// adjustl(trim(atthk_F)) //" timestep=1 "
       write(errstr,*) "*atthkgen command = "
       call message(AppMsg,Always)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Always)
       call system(trim(command))
    end if
    !
    call run_edetectchain(evtlist,imglist,piminlist,pimaxlist,ecflist,inpnoot_F)
    !
    ! Now make region files, masks, and cheese masks
    ! EMOS1
    if (mos1) then
       ! OLDNEW: mos1S001-mask-im.fits mos1S001-fovim(h/s/t)mask
       ! OLDNEW: mos1S001-exp-im.fits mos1S001-fovim(h/s/t)exp
       ! OLDNEW: mos1S001-bkg_region-det.fits mos1S001-bkgreg(h/s/t)det.fits
       ! OLDNEW: mos1S001-bkg_region-sky.fits mos1S001-bkgreg(h/s/t)sky.fits
       ! OLDNEW: mos1S001-cheese.fits mos1S001-cheese(h/s/t).fits 
       ! OLDNEW: mos1S001-mask-im-s.fits mos1S001-fovim(h/s/t)mask
       ! OLDNEW: mos1S001-exp-im-s.fits mos1S001-fovim(h/s/t)exp
       ! OLDNEW: mos1S001-bkg_region-det-s.fits mos1S001-bkgreg(h/s/t)det.fits
       ! OLDNEW: mos1S001-bkg_region-sky-s.fits mos1S001-bkgreg(h/s/t)sky.fits
       ! OLDNEW: mos1S001-cheese-s.fits mos1S001-cheese(h/s/t).fits 
       ! OLDNEW: mos1S001-mask-im-h.fits mos1S001-fovim(h/s/t)mask
       ! OLDNEW: mos1S001-exp-im-h.fits mos1S001-fovim(h/s/t)exp
       ! OLDNEW: mos1S001-bkg_region-det-h.fits mos1S001-bkgreg(h/s/t)det.fits
       ! OLDNEW: mos1S001-bkg_region-sky-h.fits mos1S001-bkgreg(h/s/t)sky.fits
       ! OLDNEW: mos1S001-cheese-h.fits mos1S001-cheese(h/s/t).fits 
       ! total band: always make (whether singlet or list)
       call makeOutName(inmos1_F,detmask_F,"t",1,"mask") ! mos1S001-fovimtmask
       call makeOutName(inmos1_F,expimg_F,"t",1,"exp")   ! mos1S001-fovimtexp
       !
       call run_emask(detmask_F,expimg_F)
       !
       call makeOutName(inmos1_F,regiondet_F,"t",2,any) ! mos1S001-bkgregtdet.fits
       call makeOutName(inmos1_F,regionsky_F,"t",3,any) ! mos1S001-bkgregtsky.fits
       call run_region(inmos1_F,regiondet_F,regionsky_F,ratet,mlmin,2,&
            dist,scale,1)
       call makeOutName(inmos1_F,cheese_F,"t",5,any) ! mos1S001-cheeset.fits
       call run_makemask(outM1fovT_F,detmask_F,regionsky_F,cheese_F)
       if (nBandsL .gt. 1) then
          ! soft band
          call makeOutName(inmos1_F,detmask_F,"s",1,"mask") ! mos1S001-fovimsmask
          call makeOutName(inmos1_F,expimg_F,"s",1,"exp")   ! mos1S001-fovimsexp
          call run_emask(detmask_F,expimg_F)
          call makeOutName(inmos1_F,regiondet_F,"s",2,any) ! mos1S001-bkgregsdet.fits
          call makeOutName(inmos1_F,regionsky_F,"s",3,any) ! mos1S001-bkgregssky.fits
          call run_region(inmos1_F,regiondet_F,regionsky_F,rates,mlmin,2,&
               dist,scale,1)
          call makeOutName(inmos1_F,cheese_F,"s",5,any) ! mos1S001-cheeses.fits
          call run_makemask(outM1fovS_F,detmask_F,regionsky_F,cheese_F)
          ! hard band
          call makeOutName(inmos1_F,detmask_F,"h",1,"mask") ! mos1S001-fovimsmask
          call makeOutName(inmos1_F,expimg_F,"h",1,"exp")   ! mos1S001-fovimsexp
          call run_emask(detmask_F,expimg_F)
          call makeOutName(inmos1_F,regiondet_F,"h",2,any) ! mos1S001-bkgreghdet.fits
          call makeOutName(inmos1_F,regionsky_F,"h",3,any) ! mos1S001-bkgreghsky.fits
          call run_region(inmos1_F,regiondet_F,regionsky_F,rateh,mlmin,2,&
               dist,scale,2)
          call makeOutName(inmos1_F,cheese_F,"h",5,any) ! mos1S001-cheeseh.fits
          call run_makemask(outM1fovH_F,detmask_F,regionsky_F,cheese_F)
       end if
    end if
    ! EMOS2
    if (mos2) then
       ! total band 
       call makeOutName(inmos2_F,detmask_F,"t",1,"mask") ! mos2S002-fovimtmask
       call makeOutName(inmos2_F,expimg_F,"t",1,"exp")   ! mos2S002-fovimtexp
       call run_emask(detmask_F,expimg_F)
       call makeOutName(inmos2_F,regiondet_F,"t",2,any) ! mos2S002-bkgregtdet.fits
       call makeOutName(inmos2_F,regionsky_F,"t",3,any) ! mos2S001-bkgregtsky.fits
       call run_region(inmos2_F,regiondet_F,regionsky_F,ratet,mlmin,3,&
            dist,scale,1)
       call makeOutName(inmos2_F,cheese_F,"t",5,any) ! mos2S002-cheeset.fits
       call run_makemask(outM2fovT_F,detmask_F,regionsky_F,cheese_F)
       if (nBandsL .gt. 1) then
          ! soft band
          call makeOutName(inmos2_F,detmask_F,"s",1,"mask") ! mos2S002-fovimsmask
          call makeOutName(inmos2_F,expimg_F,"s",1,"exp")   ! mos2S002-fovimsexp
          call run_emask(detmask_F,expimg_F)
          call makeOutName(inmos2_F,regiondet_F,"s",2,any) ! mos1S001-bkgregsdet.fits
          call makeOutName(inmos2_F,regionsky_F,"s",3,any) ! mos1S001-bkgregsdet.fits
          call run_region(inmos2_F,regiondet_F,regionsky_F,rates,mlmin,3,&
               dist,scale,1)
          call makeOutName(inmos2_F,cheese_F,"s",5,any) ! mos2S002-cheeses.fits
          call run_makemask(outM2fovS_F,detmask_F,regionsky_F,cheese_F)
          ! hard band
          call makeOutName(inmos2_F,detmask_F,"h",1,"mask") ! mos2S002-fovimhmask
          call makeOutName(inmos2_F,expimg_F,"h",1,"exp")   ! mos2S002-fovimhexp
          call run_emask(detmask_F,expimg_F)
          call makeOutName(inmos2_F,regiondet_F,"h",2,any) ! mos1S001-bkreghdet.fits
          call makeOutName(inmos2_F,regionsky_F,"h",3,any) ! mos1S001-bkreghdet.fits
          call run_region(inmos2_F,regiondet_F,regionsky_F,rateh,mlmin,3,&
               dist,scale,2)
          call makeOutName(inmos2_F,cheese_F,"h",5,any) ! mos2S002-cheeseh.fits
          call run_makemask(outM2fovH_F,detmask_F,regionsky_F,cheese_F)
       end if
    end if
    ! PN
    if (pn) then
       ! total band
       call makeOutName(inpn_F,detmask_F,"t",1,"mask") ! pnS002-fovimtmask
       call makeOutName(inpn_F,expimg_F,"t",1,"exp")   ! pnS002-fovimtexp
       call run_emask(detmask_F,expimg_F)
       call makeOutName(inpn_F,regiondet_F,"t",2,any) ! mos1S001-bkregtdet.fits
       call makeOutName(inpn_F,regionsky_F,"t",3,any) ! mos1S001-bkregtdet.fits
       call run_region(inpn_F,regiondet_F,regionsky_F,ratet,mlmin,3,&
            dist,scale,1)
       call makeOutName(inpn_F,cheese_F,"t",5,any) ! pnS003-cheeset.fits
       call run_makemask(outPNfovT_F,detmask_F,regionsky_F,cheese_F)
       if (nBandsL .gt. 1) then
          ! soft band
          call makeOutName(inpn_F,detmask_F,"s",1,"mask") ! pnS002-fovimsmask
          call makeOutName(inpn_F,expimg_F,"s",1,"exp")   ! pnS002-fovimsexp
          call run_emask(detmask_F,expimg_F)
          call makeOutName(inpn_F,regiondet_F,"s",2,any) ! mos1S001-bkregsdet.fits
          call makeOutName(inpn_F,regionsky_F,"s",3,any) ! mos1S001-bkregsdet.fits
          call run_region(inpn_F,regiondet_F,regionsky_F,rates,mlmin,3,&
               dist,scale,1)
          call makeOutName(inpn_F,cheese_F,"s",5,any) ! pnS003-cheeses.fits
          call run_makemask(outPNfovS_F,detmask_F,regionsky_F,cheese_F)
          ! hard band
          call makeOutName(inpn_F,detmask_F,"h",1,"mask") ! pnS002-fovimhmask
          call makeOutName(inpn_F,expimg_F,"h",1,"exp")   ! pnS002-fovimhexp
          call run_emask(detmask_F,expimg_F)
          call makeOutName(inpn_F,regiondet_F,"h",2,any) ! mos1S001-bkreghdet.fits
          call makeOutName(inpn_F,regionsky_F,"h",3,any) ! mos1S001-bkreghdet.fits
          call run_region(inpn_F,regiondet_F,regionsky_F,rateh,mlmin,3,&
               dist,scale,2)
          call makeOutName(inpn_F,cheese_F,"h",5,any) ! pnS003-cheeseh.fits
          call run_makemask(outPNfovH_F,detmask_F,regionsky_F,cheese_F)
       end if
    end if
    ! which of these can be deleted???!!!
    ! cheese makes these files (plus s and h versions if bands selected):
    ! pnS003-bkregtsky.fits, pnS003-cheeset.fits, mos2S002-cheeset.fits,
    ! pnS003-bkregtdet.fits,pnS003-fovimtmask.fits,mos1S001-bkregtsky.fits,
    ! mos1S001-cheeset.fits,mos2S002-bkregtdet.fits,mos2S002-bkregtsky.fits,
    ! mos2S002-fovimtmask.fits,mos1S001-bkregtdet.fits,mos1S001-fovimtmask.fits,
    ! mos2S002-fovimtsen.fits,mos1S001-fovimtsen.fits,pnS003-fovimtsen.fits,
    ! emllist.fits,mos1S001-fovimtsmap.fits,mos2S002-fovimtsmap.fits,
    ! pnS003-fovimtsmap.fits,eboxlist_m.fits,mos2S002-fovimtbkg.fits,
    ! mos1S001-fovimtbkg.fits,pnS003-fovimtbkg.fits,eboxlist_l.fits,
    ! mos2S002-fovimtexp.fits,mos1S001-fovimtexp.fits,pnS003-fovimtexp.fits,
    ! atthk.fits,filtered.fits,pnS003-fovimt.fits,mos1S001-fovimt.fits,
    ! mos2S002-fovimt.fits
    !
    if (.not.keep) then
       ! add actual files to kill here ???!!!
       write(errstr,*) "CHEESE: keepinterfiles NOT set"
       call message(AppMsg,Always)
       write(errstr,*) "CHEESE: Deleting intermediary files..."
       call message(AppMsg,Always)
       !write(errstr,*) "Object image: ",trim(adjustl(objimg_F))
       !call message(AppMsg,Always)
       !call unlink(trim(adjustl(objimg_F)))
    end if

    return

  end subroutine cheese

  subroutine makeOutName(infile,outfile,band,type,any)
    !
    ! create output file name based on input eventlist
    ! 
    type(DataSetT) :: in_S
    type(TableT)   :: in_T
    type(ArrayT)   :: in_A
    character(len=*), intent(in) :: any
    character(len=256), intent(in) :: infile
    character(len=1024), intent(out) :: outfile
    character(len=25) :: smode
    character(len=10) :: obsid, expid, brand
    character(len=6) :: instr, filter, coord, suffix
    character(len=2) :: inst
    character(len=1) :: band
    !
    integer :: typeF, status, revolut, nRows, type, frmtime, &
         elow, ehigh
    logical :: warndie
    real :: back, area, expo
    suffix  = '.fits' ! default case
    coord   = ''
    brand   = ''
    outfile = ''
    select case (type)
    case(1) ! fov image in sky coords
       brand   = "-fovim"
    case(2) ! region file in det coords
       brand   = "-bkgreg" ! bkregtreg
       coord  = "det" 
    case(3) ! region file in sky coords
       brand   = "-bkgreg" ! bkregtreg
       coord  = "sky" 
    case(4) ! emask file
       brand = "-fovexp"
    case(5) ! final cheese mask
       brand = "-cheese"
    case default
       brand = ''
    end select
    write(errstr,*) &
         "CHEESE: Reading info from this file to make new name: ",trim(infile)
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(infile),in_S,1,status,warndie)
    !
    ! Get critical header keywords 
    !
    status = 0
    warndie = .TRUE.
    typeF = 1 ! 1 = EVLI, 2 = SPEC, 3 = IMG, most of these are dummies
    call read_header_info(typeF,in_S,in_T,in_A,inst,obsid,expid,&
         smode,frmtime,filter,revolut,back,area,expo,elow,ehigh,&
         nRows,warndie)
    select case (inst)
    case('M1') 
       instr = "mos1"
    case('M2')
       instr = "mos2"
    case('PN')
       instr = "pn"
    case default
       instr = ""
    end select
    write(errstr,*) "CHEESE: Old (input) file     : ",trim(infile)
    call message(AppMsg,Always)
    outfile = trim(adjustl(instr)) // trim(adjustl(expid)) // &
         trim(adjustl(brand)) // trim(adjustl(band)) // &
         trim(adjustl(coord)) // trim(adjustl(any)) // &
         trim(adjustl(suffix))
    write(errstr,*) "CHEESE: New (output) filename: ",trim(adjustl(outfile))
    call message(AppMsg,Always)
    !    print*,'instr = ',instr
    !    print*,'expid = ',expid
    !    print*,'brand = ',brand
    !    print*,'band  = ',band
    !    print*,'coord = ',coord
    !    print*,'any   = ',any
    !    print*,'suffix = ',suffix
    !    print*,'outfile = ',trim(adjustl(outfile))
    return

  end subroutine makeOutName
  !
  subroutine makeImFile(infile,outfile,inst,celow,cehigh)
    !
    ! Make the clean FOV image files in sky coords
    !
    character(len=*), intent(inout) :: infile, outfile
    character(len=*), intent(in) :: celow, cehigh
    character(len=2) :: inst
    character(len=4096) :: command
    integer :: nMOS1, nMOS2, nPN, elow, ehigh
    logical :: die
    if ((inst .eq. 'M1') .or. (inst .eq. 'M2')) then 
       write(errstr,*) "Checking FOV sky coord image files for " // inst
       call message(AppMsg,Always)
       die = .FALSE. ! fix all of these check_clobbers???!!!
       call check_clobber(trim(adjustl(outfile)),die)
       command =  "evselect table="// trim(adjustl(infile)) //":EVENTS " // &
            "withfilteredset=yes expression='(PATTERN<=12)&&(#XMMEA_EM)&&((FLAG & 0x766aa000) == 0)&&" // &
            "(PI in [" // trim(adjustl(celow)) // ":" // trim(adjustl(cehigh)) // &
            "])' " // "filtertype=expression imagebinning='imageSize' " // &
            "imagedatatype='Int32' imageset=" // trim(adjustl(outfile)) // &
            " squarepixels=yes ignorelegallimits=yes "// &
            "withxranges=yes withyranges=yes xcolumn='X' ximagesize=900 " // &
            "ximagemax=48400 ximagemin=3401 ycolumn='Y' yimagesize=900 " // &
            "yimagemax=48400 yimagemin=3401 updateexposure=yes filterexposure=yes " // &
            "verbosity=4"
       write(errstr,*) inst // " Flare free FOV image evselect command = "
       call message(AppMsg,Always)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Always)
       call system(trim(command))
       !
    elseif (inst .eq. 'PN') then 
       write(errstr,*) "Checking FOV sky coord image files for PN"
       die = .FALSE.
       call check_clobber(trim(adjustl(outfile)),die)
       command = "evselect table="// trim(adjustl(infile)) // ":EVENTS " //&
            "withfilteredset=yes expression='(PATTERN<=4)&&((FLAG & 0xb0000) == 0)&&((FLAG & 0x762a097c) == 0)&&" //&
            "(PI in [" // trim(adjustl(celow)) // ":" // trim(adjustl(cehigh)) // &
            "])&&" // "(DETY in [-16510:14345])' filtertype=expression " //& 
            "imagebinning='imageSize' imagedatatype='Int32' " //&
            "imageset=" // trim(adjustl(outfile)) //&
            " ignorelegallimits=yes withxranges=yes withyranges=yes " //&
            "xcolumn='X' ximagesize=900 ximagemax=48400 ximagemin=3401 " //&
            "ycolumn='Y' yimagesize=900 yimagemax=48400 yimagemin=3401 " //& 
            "updateexposure=yes filterexposure=yes verbosity=4"
       write(errstr,*) inst // " Flare free FOV image evselect command = "
       call message(AppMsg,Always)
       write(errstr,*) trim(adjustl(command))
       call message(AppMsg,Always)
       call system(trim(adjustl(command)))
    else
       call error("notM1M2PN","Instrument is not M1, M2, or PN, weirdly")
    end if


    return

  end subroutine makeImFile
  !
  !
  !
  subroutine run_edetectchain(evtlist,imglist,piminlist,pimaxlist,ecflist,pnoot_F)
    !
    ! routine to run SAS task edetect_chain
    !
    character(len=1024), intent(in) :: imglist, evtlist ! fix to match max list value???!!!
    character(len=4096) :: command
    character(len=256), intent(in) :: piminlist, pimaxlist, ecflist, pnoot_F
    character(len=12) :: emllist_F = "emllist.fits"
    logical :: die
    die = .FALSE.
    call check_clobber(trim(adjustl(emllist_F)),die)
    !
    if (pnoot_F .ne. "") then
       command = "edetect_chain imagesets='" // trim(adjustl(imglist)) // &
            " ' eventsets='" // trim(adjustl(evtlist)) // &
            " ' attitudeset=atthk.fits pimin='" // trim(adjustl(piminlist)) // &
            " ' pimax='" // trim(adjustl(pimaxlist)) // &
            " ' witheexpmap=yes likemin=10 ecf='" // trim(adjustl(ecflist)) // &
            " ' eboxl_likemin=15 eboxm_likemin=15 ebox_withdetmask=yes "// &
            " ebox_withexpimage=yes esp_scut=0.01 esp_withootset=yes " // &
            " esp_ooteventset='" // trim(adjustl(pnoot_F)) // "' esp_nsplinenodes=20" // &
            " eml_ecut=15 eml_scut=15 eml_withtwostage=yes eml_withthreshold=yes " // &
            " eml_threshold=15 esen_mlmin=10 eml_list='emllist.fits'"
    else
       command = "edetect_chain imagesets='" // trim(adjustl(imglist)) // &
            " ' eventsets='" // trim(adjustl(evtlist)) // &
            " ' attitudeset=atthk.fits pimin='" // trim(adjustl(piminlist)) // &
            " ' pimax='" // trim(adjustl(pimaxlist)) // &
            " ' witheexpmap=yes likemin=10 ecf='" // trim(adjustl(ecflist)) // &
            " ' eboxl_likemin=15 eboxm_likemin=15 ebox_withdetmask=yes "// &
            " ebox_withexpimage=yes esp_scut=0.01 esp_withootset=no " // &
            " esp_nsplinenodes=20" // &
            " eml_ecut=15 eml_scut=15 eml_withtwostage=yes eml_withthreshold=yes " // &
            " eml_threshold=15 esen_mlmin=10 eml_list='emllist.fits'"
    end if

    write(errstr,*) "*edetect_chain command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(command))
    return

  end subroutine run_edetectchain

  subroutine StripSpaces(string)
    character(len=*) :: string
    integer :: stringLen 
    integer :: last, actual

    stringLen = len (string)
    last = 1
    actual = 1

    do while (actual < stringLen)
       if (string(last:last) == ' ') then
          actual = actual + 1
          string(last:last) = string(actual:actual)
          string(actual:actual) = ' '
       else
          last = last + 1
          if (actual < last) actual = last
       endif
    end do

  end subroutine StripSpaces
  !
  !
  !
  subroutine run_region(infile,detfile,skyfile,rate,mlmin,det,dist,scale,idband)
    !
    ! Make background region files
    !
    character(len=*), intent(in) :: infile, detfile, skyfile
    character(len=10) :: cdist, crate, cdet, cmlmin, cscale, cidband
    character(len=4096) :: command
    integer :: det, idband
    real :: rate, mlmin, dist, scale

    ! formats taken from parameter constraints. 
    write(cdet,'(I1)') det
    write(cidband,'(I1)') idband
    write(cdist,'(F7.4)') dist
    write(crate,'(F7.3)') rate 
    write(cmlmin,'(F7.4)') mlmin
    write(cscale,'(F7.4)') scale
    crate = trim(adjustl(crate)) // "e-14"
    !           
    command = "region eventset=" // trim(adjustl(infile)) // &
         " operationstyle=global srclisttab=emllist.fits:SRCLIST " // & 
         "expression='(DIST_NN >= " // trim(adjustl(cdist)) // &
         ")&&(FLUX >= " //  trim(adjustl(crate)) // ")" // &
         "&&(ID_INST == " // trim(adjustl(cdet)) // ")&&(DET_ML >= " // &
         trim(adjustl(cmlmin)) // ")&&(ID_BAND == " // trim(adjustl(cidband)) // &
         ")' bkgregionset=" // trim(adjustl(detfile)) // &
         " bkgfraction=" // trim(adjustl(cscale)) // &
         " radiusstyle=contour nosrcellipse=no " // & 
         "outunit=detxy verbosity=4"
    write(errstr,*) " Running region in DET coords command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(command))
    command = "region eventset=" // trim(adjustl(infile)) // &
         " operationstyle=global srclisttab=emllist.fits:SRCLIST " // & 
         "expression='(DIST_NN >= " // trim(adjustl(cdist)) // &
         ")&&(FLUX >= " //  trim(adjustl(crate)) // ")" // &
         "&&(ID_INST == " // trim(adjustl(cdet)) // ")&&(DET_ML >= " // &
         trim(adjustl(cmlmin)) // ")&&(ID_BAND == " // trim(adjustl(cidband)) // &
         ")' bkgregionset=" // trim(adjustl(skyfile)) // &
         " bkgfraction=" // trim(adjustl(cscale)) // &
         " radiusstyle=contour nosrcellipse=no " // & 
         "outunit=xy verbosity=4"
    write(errstr,*) " Running region in SKY coords command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(command))
    !    
    return

  end subroutine run_region

  subroutine run_emask(infile,maskfile)
    !
    ! Make masks files
    !
    character(len=*), intent(in) :: infile, maskfile
    character(len=4096) :: command
    command = "emask detmaskset=" // trim(adjustl(infile)) // & 
         " expimageset=" // trim(adjustl(maskfile)) // & 
         " threshold1=0.1 threshold2=0.5 verbosity=4"
    write(errstr,*) " YYYYY Running emask command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(command))

    return

  end subroutine run_emask
  !
  !
  !
  subroutine run_makemask(imagefile,maskfile,regionfile,cheesefile)
    !
    ! Make cheese masks files
    ! imagefile : old=mos1S001-obj-im.fits  new=mos1S001-fovimt.fits
    ! maskfile  : old=mos1S001-mask-im.fits new=mos1S001-fovimtmask.fits
    ! cheesefile: old=mos1S001-cheese.fits  new=mos1S001-fovimtcheese.fits
    ! regionfile: old=mos1S001-bkg_region-sky.fits new=mos1S001-bkgregsky.fits
    !
    character(len=*), intent(in) :: imagefile, maskfile, cheesefile, regionfile
    character(len=4096) :: command
    ! in
    command = "makemask imagefile=" // trim(adjustl(imagefile)) // &
         " maskfile=" // trim(adjustl(maskfile)) // &
         " cheesefile=" // trim(adjustl(cheesefile)) // &
         " regionfile=" // trim(adjustl(regionfile))

    write(errstr,*) " Running makemask command = "
    call message(AppMsg,Always)
    write(errstr,*) trim(adjustl(command))
    call message(AppMsg,Always)
    call system(trim(command))

    return

  end subroutine run_makemask
  !
  !
  !
  subroutine printparams (mos1_F,mos2_F,pn_F,pnoot_F,scale,rateT,rateS,rateH,dist,mlmin,keep)

    character(len=*) :: mos1_F, mos2_F, pn_F, pnoot_F
    real :: scale, rateT, rateS, rateH, dist, mlmin
    logical :: keep
    if (mos1_F .ne. "") then
       write(errstr,*) "MOS1 input file   : ",trim(adjustl(mos1_F))
       call message(AppMsg,Always)
    endif
    if (mos2_F .ne. "") then
       write(errstr,*) "MOS2 input file   : ",trim(adjustl(mos2_F))
       call message(AppMsg,Always)
    endif
    if (pn_F .ne. "") then
       write(errstr,*) "PN input file     :   ",trim(adjustl(pn_F))
       call message(AppMsg,Always)
    endif
    if (pnoot_F .ne. "") then
       write(errstr,*) "PN OOT input file :   ",trim(adjustl(pnoot_F))
       call message(AppMsg,Always)
    endif
    !
    write(errstr,*) "Energy Fraction (exclusion radius for point sources):     ",scale
    call message(AppMsg,Always)
    write(errstr,*) "Flux threshold for exclusion of point sources (fullband): ",rateT
    call message(AppMsg,Always)
    write(errstr,*) "Flux threshold for exclusion of point sources (softband): ",rateS
    call message(AppMsg,Always)
    write(errstr,*) "Flux threshold for exclusion of point sources (hardband): ",rateH
    call message(AppMsg,Always)
    write(errstr,*) "Minimum separation (arcsec) between masked sources:      ",dist
    call message(AppMsg,Always)
    write(errstr,*) "Minimum Maximum Likelihood detection value:               ",mlmin
    call message(AppMsg,Always)
    write(errstr,*) "Keep intermediary files?: ",keep
    call message(AppMsg,Always)
    return

  end subroutine printparams
  !
  subroutine chuck_clobber(outfile,die,go)

    character(len=*) :: outfile
    logical :: exists, die, go
    !                                                                                                              
    ! In any ESAS task, if clobber is not set, die (early!)                                                        
    !
    go = .FALSE.
    inquire(FILE=trim(outfile),EXIST=exists)
    if (exists) then
       if (clobber()) then
          write(errstr,*)"Output file, " // trim(adjustl(outfile)) // &
               " exists and clobber IS set. Deleting file, creating new version."
          call message(AppMsg,Always)
          call unlink (trim(adjustl(outfile)))
          go = .TRUE.
       else
          if(die) then
             write(errstr,*)"Output file, " // trim(adjustl(outfile)) // &
                  " exists and clobber is NOT set, exiting."
             call error("noClobber",trim(adjustl(errstr)))
          else
             write(errstr,*)"Output file, " // trim(adjustl(outfile)) // &
                  " exists and clobber is NOT set, will use existing version!"
             call warning("noClobber",trim(adjustl(errstr)))
          endif
       endif
    end if

    return

  end subroutine chuck_clobber

end module cheese_mod
